package registry // import "github.com/docker/docker/registry"

import (
	"crypto/tls"
	"net/url"
	"strings"

	registrytypes "github.com/docker/docker/api/types/registry"
	"github.com/docker/go-connections/tlsconfig"
)

func (s *DefaultService) lookupV2Endpoints(hostname string) (endpoints []APIEndpoint, err error) {
	var tlsConfig *tls.Config

	// if s.config.Registries is set, lookup regsitry mirror addr from s.config.Registries
	if len(s.config.Registries) > 0 {
		reg := s.config.Registries.FindRegistry(hostname)

		if reg != nil {
			var regEndpoints []registrytypes.Endpoint = reg.Mirrors

			for _, regEP := range regEndpoints {
				official := regEP.Address == registrytypes.DefaultEndpoint.Address
				regURL := regEP.GetURL()

				if official {
					tlsConfig = tlsconfig.ServerDefault()
				} else {
					tlsConfig, err = s.tlsConfigForMirror(regURL)
					if err != nil {
						return nil, err
					}
				}
				tlsConfig.InsecureSkipVerify = regEP.InsecureSkipVerify
				endpoints = append(endpoints, APIEndpoint{
					URL:          regURL,
					Version:      APIVersion2,
					Official:     official,
					TrimHostname: true,
					TLSConfig:    tlsConfig,
					Prefix:       hostname,
					// the last endpoint is not considered a mirror
					Mirror: true,
				})
			}
			// don't return here, otherwise the hostname itself will not be appended to the endpoints,
			// and the hostname itself will not be tried, which is not a desired action.
		}
	}
	if hostname == DefaultNamespace || hostname == IndexHostname {
		tlsConfig = tlsconfig.ServerDefault()
		// v2 mirrors
		for _, mirror := range s.config.Mirrors {
			if !strings.HasPrefix(mirror, "http://") && !strings.HasPrefix(mirror, "https://") {
				mirror = "https://" + mirror
			}
			mirrorURL, err := url.Parse(mirror)
			if err != nil {
				return nil, err
			}
			mirrorTLSConfig, err := s.tlsConfigForMirror(mirrorURL)
			if err != nil {
				return nil, err
			}
			endpoints = append(endpoints, APIEndpoint{
				URL: mirrorURL,
				// guess mirrors are v2
				Version:      APIVersion2,
				Mirror:       true,
				TrimHostname: true,
				TLSConfig:    mirrorTLSConfig,
			})
		}
		// v2 registry
		endpoints = append(endpoints, APIEndpoint{
			URL:          DefaultV2Registry,
			Version:      APIVersion2,
			Official:     true,
			TrimHostname: true,
			TLSConfig:    tlsConfig,
		})

		return endpoints, nil
	}
	ana := allowNondistributableArtifacts(s.config, hostname)

	tlsConfig, err = s.tlsConfig(hostname)
	if err != nil {
		return nil, err
	}

	endpoints = append(endpoints, APIEndpoint{
		URL: &url.URL{
			Scheme: "https",
			Host:   hostname,
		},
		Version:                        APIVersion2,
		AllowNondistributableArtifacts: ana,
		TrimHostname:                   true,
		TLSConfig:                      tlsConfig,
	})

	if tlsConfig.InsecureSkipVerify {
		endpoints = append(endpoints, APIEndpoint{
			URL: &url.URL{
				Scheme: "http",
				Host:   hostname,
			},
			Version:                        APIVersion2,
			AllowNondistributableArtifacts: ana,
			TrimHostname:                   true,
			// used to check if supposed to be secure via InsecureSkipVerify
			TLSConfig: tlsConfig,
		})
	}

	return endpoints, nil
}
